# coding: utf-8
import base64
import json
import random
import string
import sys
import time
from six.moves.urllib.parse import urlencode
import webbrowser

from tccli import oauth
from tccli.plugins.auth import texts

_APP_ID = 100038427476
_AUTH_URL = "https://cloud.tencent.com/open/authorize"
_REDIRECT_URL = "https://cli.cloud.tencent.com/oauth"
_SITE = "cn"
_DEFAULT_LANG = "zh-CN"

_START_SEARCH_PORT = 9000
_END_SEARCH_PORT = _START_SEARCH_PORT + 100


def print_message(msg):
    print(msg)
    sys.stdout.flush()


def login_command_entrypoint(args, parsed_globals):
    language = parsed_globals.get("language")
    if not language:
        language = _DEFAULT_LANG
    texts.set_lang(language)

    profile = parsed_globals.get("profile", "default")
    if not profile:
        profile = "default"

    browser = args.get("browser")

    login(browser != "no", profile, language)


def login(use_browser, profile, language):
    characters = string.ascii_letters + string.digits
    state = ''.join(random.choice(characters) for _ in range(10))

    if use_browser:
        token, cred = _get_token(state, language)
    else:
        token, cred = _get_token_no_browser(state, language)

    if token["state"] != state:
        raise ValueError("invalid state %s" % token["state"])

    oauth.save_credential(token, cred, profile)

    print_message("")
    print_message(texts.get("login_success") % oauth.cred_path_of_profile(profile))


def _get_token(state, language):
    from tccli.plugins.auth import browser_flow

    port, result_queue = browser_flow.try_run(_START_SEARCH_PORT, _END_SEARCH_PORT)

    redirect_params = {
        "redirect_url": "http://localhost:%d" % port,
        "lang": language,
        "site": _SITE,
    }
    redirect_query = urlencode(redirect_params)
    redirect_url = _REDIRECT_URL + "?" + redirect_query
    url_params = {
        "scope": "login",
        "app_id": _APP_ID,
        "redirect_url": redirect_url,
        "state": state,
    }
    url_query = urlencode(url_params)
    auth_url = _AUTH_URL + "?" + url_query

    if not webbrowser.open(auth_url):
        print_message(texts.get("login_failed_due_to_no_browser"))
        sys.exit(1)

    print_message(texts.get("login_prompt"))
    print_message(auth_url)

    # use polling to avoid being unresponsive in python2
    while result_queue.empty():
        time.sleep(1)

    result = result_queue.get()
    if isinstance(result, Exception):
        raise result

    return result


def _get_token_no_browser(state, language):
    redirect_params = {
        "browser": "no",
        "lang": language,
        "site": _SITE,
    }
    redirect_query = urlencode(redirect_params)
    redirect_url = _REDIRECT_URL + "?" + redirect_query
    url_params = {
        "scope": "login",
        "app_id": _APP_ID,
        "redirect_url": redirect_url,
        "state": state,
    }
    url_query = urlencode(url_params)
    auth_url = _AUTH_URL + "?" + url_query

    print_message(texts.get("login_prompt_no_browser"))
    print_message("")
    print_message(auth_url)

    try:
        input_func = raw_input
    except NameError:
        input_func = input

    user_input = input_func(texts.get("login_prompt_code_no_browser"))
    token = json.loads(base64.b64decode(user_input))
    cred = oauth.get_temp_cred(token["accessToken"], token["site"])
    return token, cred



import time
import traceback
import socket
from threading import Thread
from tccli import oauth

try:
    from urlparse import urlparse, parse_qs
    from Queue import Queue
    from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
    from SocketServer import ThreadingTCPServer, TCPServer
except ImportError:
    from urllib.parse import urlparse, parse_qs
    from queue import Queue
    from http.server import HTTPServer, BaseHTTPRequestHandler
    from socketserver import ThreadingTCPServer, TCPServer


# chrome keeps previous connection alive, so use threading to avoid blocking
class ThreadingHTTPServer(ThreadingTCPServer):
    allow_reuse_address = 1

    def server_bind(self):
        """Override server_bind to store the server name."""
        TCPServer.server_bind(self)
        host, port = self.socket.getsockname()[:2]
        self.server_name = socket.getfqdn(host)
        self.server_port = port


class HTTPHandler(BaseHTTPRequestHandler):
    result_queue = Queue(1)

    def do_GET(self):
        try:
            parsed_url = urlparse(self.path)
            query_vals = parse_qs(parsed_url.query)

            open_id = query_vals.get("open_id")[0]
            access_token = query_vals.get("access_token")[0]
            refresh_token = query_vals.get("refresh_token")[0]
            expires_at = int(query_vals.get("expires_at")[0])
            state = query_vals.get("state")[0]
            redirect_url = query_vals.get("redirect_url")[0]
            site = query_vals.get("site")[0]
            token = {
                "openId": open_id,
                "accessToken": access_token,
                "refreshToken": refresh_token,
                "expiresAt": expires_at,
                "state": state,
                "site": site,
            }
            cred = oauth.get_temp_cred(token["accessToken"], token["site"])
            self.result_queue.put((token, cred))
            self.send_response(307)
            self.send_header("Location", redirect_url)
            self.end_headers()
        except Exception:
            err = traceback.format_exc()
            print(err)
            self.send_response(400)
            self.end_headers()
            self.wfile.write("login failed due to the following error:\n\n".encode("utf-8"))
            self.wfile.write(err.encode("utf-8"))
            self.wfile.flush()

    # suppress debug message
    def log_message(self, format, *args):
        pass


def try_run(start_search_port, end_search_port):
    port = start_search_port

    while port <= end_search_port:
        server_address = ('', port)
        try:
            ThreadingHTTPServer.daemon_threads = True
            httpd = ThreadingHTTPServer(server_address, HTTPHandler)
            t = Thread(target=httpd.serve_forever)
            t.setDaemon(True)
            t.start()
            return port, HTTPHandler.result_queue
        except socket.error:
            port += 1

    raise socket.error("no port available from range [%d, %d]" % (start_search_port, end_search_port))



# coding: utf-8
import os

from tccli import oauth
from tccli.plugins.auth import texts


def logout_command_entrypoint(args, parsed_globals):
    language = parsed_globals.get("language")
    if not language:
        language = "zh-CN"
    texts.set_lang(language)

    profile = parsed_globals.get("profile", "default")
    if not profile:
        profile = "default"

    cred_path = oauth.cred_path_of_profile(profile)
    if os.path.exists(cred_path):
        os.remove(cred_path)
    print(texts.get("logout") % cred_path)



# encoding: utf-8
from tccli.plugins.auth.login import login_command_entrypoint
from tccli.plugins.auth.logout import logout_command_entrypoint

service_name = "auth"
service_version = "2024-08-20"

_spec = {
    "metadata": {
        "serviceShortName": service_name,
        "apiVersion": service_version,
        "description": "auth related commands",
    },
    "actions": {
        "login": {
            "name": "登陆",
            "document": "login through sso",
            "input": "loginRequest",
            "output": "loginResponse",
            "action_caller": login_command_entrypoint,
        },
        "logout": {
            "name": "登出",
            "document": "remove local credential file",
            "input": "logoutRequest",
            "output": "logoutResponse",
            "action_caller": logout_command_entrypoint,
        },
    },
    "objects": {
        "loginRequest": {
            "members": [
                {
                    "name": "browser",
                    "member": "string",
                    "type": "string",
                    "required": False,
                    "document": "use browser=no to indicate no browser login mode",
                },
            ],
        },
        "loginResponse": {
            "members": [],
        },
        "logoutRequest": {
            "members": [],
        },
        "logoutResponse": {
            "members": [],
        },
    },
    "version": "1.0",
}


def register_service(specs):
    specs[service_name] = {
        service_version: _spec,
    }



import json
import os
import time
import uuid

import requests

_API_ENDPOINT = "https://cli.cloud.tencent.com"
_CRED_REFRESH_SAFE_DUR = 60 * 5
_SKEY_REFRESH_SAFE_DUR = 3600 * 12 - 300


def maybe_refresh_credential(profile):
    cred_path = cred_path_of_profile(profile)
    try:
        with open(cred_path, "r") as cred_file:
            cred = json.load(cred_file)
    except IOError:
        # file not found, don't check
        return

    if cred.get("type") != "sso":
        return

    try:
        now = time.time()

        expires_at = cred["expiresAt"]
        if expires_at - now > _CRED_REFRESH_SAFE_DUR:
            return

        sso_info = cred["sso"]
        site = sso_info["site"]
        sso_expires = sso_info["expiresAt"]
        if sso_expires - now < _SKEY_REFRESH_SAFE_DUR:
            # sso can't be refreshed if expired, re-login is required
            return

        saml_resp = gen_saml_response(
            sso_info["token"], "RoleSAML", sso_info["uin"], "", sso_info["roleConfigurationId"],
            sso_info["site"]
        )

        role_arn = "qcs::cam::uin/%s:roleName/TencentCloudSSO-%s" % (sso_info["uin"], sso_info["roleConfigurationName"])
        principal_arn = "qcs::cam::uin/%s:saml-provider/TencentReservedSSO-%s" % (sso_info["uin"], sso_info["zoneId"])
        cred = assume_role_with_saml(
            saml_resp["SAMLResponse"], principal_arn, role_arn, "ses-%s" % uuid.uuid4(), 7200, site)
        save_credential(cred, sso_info, profile)

    except KeyError as e:
        print("failed to refresh credential, your credential file(%s) is corrupted, %s" % (cred_path, e))

    except Exception as e:
        print("failed to refresh credential, %s" % e)


def verify_login_skey(token, site):
    api_endpoint = _API_ENDPOINT + "/api/org/verify_login_skey"

    body = {
        "TraceId": str(uuid.uuid4()),
        "Site": site,
        "LoginSkey": token,
    }
    http_response = requests.post(api_endpoint, json=body)
    try:
        resp = http_response.json()
        if "Error" in resp:
            raise ValueError(http_response.content)
        return resp
    except Exception:
        print(http_response.content)
        raise


def list_accounts_for_access_assignment(token, site):
    api_endpoint = _API_ENDPOINT + "/api/org/list_accounts_for_access_assignment"
    page_num = 1
    page_size = 20

    accounts = []
    while True:
        body = {
            "TraceId": str(uuid.uuid4()),
            "Site": site,
            "LoginToken": token,
            "PageNum": page_num,
            "PageSize": page_size,
        }
        http_response = requests.post(api_endpoint, json=body)
        resp = http_response.json()
        if "Error" in resp:
            raise ValueError(http_response.content)

        try:
            accounts.extend(resp.get("Accounts", []))
            if resp["TotalCounts"] <= len(accounts):
                break

            page_num += 1
        except Exception:
            print(http_response.content)
            raise

    return accounts


def list_role_configurations_for_account(uin, token, site):
    api_endpoint = _API_ENDPOINT + "/api/org/list_role_configurations_for_account"
    page_size = 20

    configs = []
    next_token = ""
    while True:
        body = {
            "TraceId": str(uuid.uuid4()),
            "Site": site,
            "TargetUin": uin,
            "LoginToken": token,
            "MaxResults": page_size,
            "NextToken": next_token,
        }
        http_response = requests.post(api_endpoint, json=body)
        try:
            resp = http_response.json()

            if "Error" in resp:
                raise ValueError(http_response.content)

            configs.extend(resp.get("RoleConfigurationsForAccount", []))
            if not resp["IsTruncated"]:
                break

            next_token = resp["NextToken"]
        except Exception:
            print(http_response.content)
            raise

    return configs


def gen_saml_response(token, login_type, uin, user_id, conf_id, site):
    api_endpoint = _API_ENDPOINT + "/api/org/gen_saml_response"

    body = {
        "TraceId": str(uuid.uuid4()),
        "Site": site,
        "LoginToken": token,
        "LoginType": login_type,
        "TargetUin": uin,
        "UserId": user_id,
        "RoleConfigurationId": conf_id,
    }
    http_response = requests.post(api_endpoint, json=body)
    try:
        resp = http_response.json()
        if "Error" in resp:
            raise ValueError(http_response.content)

        return resp
    except Exception:
        print(http_response.content)
        raise


def assume_role_with_saml(saml_assertion, principal_arn, role_arn, role_ses_name, dur, site):
    api_endpoint = _API_ENDPOINT + "/api/sts/assume_role_with_saml"

    body = {
        "TraceId": str(uuid.uuid4()),
        "Site": site,
        "SAMLAssertion": saml_assertion,
        "PrincipalArn": principal_arn,
        "RoleArn": role_arn,
        "RoleSessionName": role_ses_name,
        "DurationSeconds": dur,
    }
    http_response = requests.post(api_endpoint, json=body)
    try:
        resp = http_response.json()
        if "Error" in resp:
            raise ValueError(http_response.content)
        return resp
    except Exception:
        print(http_response.content)
        raise


def save_credential(cred, sso_info, profile):
    cred_path = cred_path_of_profile(profile)

    cred = {
        "type": "sso",
        "secretId": cred["Credentials"]["TmpSecretId"],
        "secretKey": cred["Credentials"]["TmpSecretKey"],
        "token": cred["Credentials"]["Token"],
        "expiresAt": cred["ExpiredTime"],
        "sso": {
            "token": sso_info["token"],
            "uin": sso_info["uin"],
            "roleConfigurationId": sso_info["roleConfigurationId"],
            "roleConfigurationName": sso_info["roleConfigurationName"],
            "zoneId": sso_info["zoneId"],
            "site": sso_info["site"],
            "authUrl": sso_info["authUrl"],
            "expiresAt": sso_info["expiresAt"],
        },
    }
    with open(cred_path, "w") as cred_file:
        json.dump(cred, cred_file, indent=4)


def cred_path_of_profile(profile):
    return os.path.join(os.path.expanduser("~"), ".tccli", profile + ".credential")


def check_login_state(state):
    api_endpoint = _API_ENDPOINT + "/api/sso/check_login_state"

    body = {
        "TraceId": str(uuid.uuid4()),
        "State": state,
    }
    http_response = requests.post(api_endpoint, json=body)
    try:
        return http_response.json()
    except Exception:
        print(http_response.content)
        raise
